///////////////////////////////////////////////////////////////////////////////
// File: split_merge.cpp                                                     //
// Description: source file for splitting/merging (contains the CJet class)  //
// This file is part of the SISCone project.                                 //
// For more details, see http://www.lpthe.jussieu.fr/~soyez/siscone          //
//                                                                           //
// Copyright (c) 2006 Gavin Salam and Gregory Soyez                          //
//                                                                           //
// This program is free software; you can redistribute it and/or modify      //
// it under the terms of the GNU General Public License as published by      //
// the Free Software Foundation; either version 2 of the License, or         //
// (at your option) any later version.                                       //
//                                                                           //
// This program is distributed in the hope that it will be useful,           //
// but WITHOUT ANY WARRANTY; without even the implied warranty of            //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             //
// GNU General Public License for more details.                              //
//                                                                           //
// You should have received a copy of the GNU General Public License         //
// along with this program; if not, write to the Free Software               //
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA //
//                                                                           //
// $Revision:: 137                                                          $//
// $Date:: 2007-01-22 17:16:47 +0100 (Mon, 22 Jan 2007)                     $//
///////////////////////////////////////////////////////////////////////////////

#include "split_merge.h"
#include "momentum.h"
#include <math.h>
#include <iostream>
#include <algorithm>

namespace siscone{

using namespace std;

/********************************************************
 * class Cjet implementation                            *
 * real Jet information.                                *
 * This class contains information for one single jet.  *
 * That is, first, its momentum carrying information    *
 * about its centre and pT, and second, its particle    *
 * content                                              *
 ********************************************************/
// default ctor
//--------------
Cjet::Cjet(){
  n = 0;
  v = Cmomentum();
}

// default dtor
//--------------
Cjet::~Cjet(){

}

// ordering jets (by pT)
//-----------------------
bool operator <(const Cjet &jet1, const Cjet &jet2){
  return jet1.v.perp2() > jet2.v.perp2();
}


/********************************************************
 * class Csplit_merge implementation                    *
 * Class used to split and merge jets.                  *
 ********************************************************/
// default ctor
//--------------
Csplit_merge::Csplit_merge(){
  merge_identical_protocones = false;
#ifdef ALLOW_MERGE_IDENTICAL_PROTOCONES
#ifdef MERGE_IDENTICAL_PROTOCONES_DEFAULT_TRUE
  merge_identical_protocones = true;
#endif
#endif
  indices = NULL;
}


// default dtor
//--------------
Csplit_merge::~Csplit_merge(){
  full_clear();
}


// initialisation function
//  - _particles  list of particles
//  - protocones  list of protocones (initial jet candidates)
//  - R2          cone radius (squared)
//-------------------------------------------------------------
int Csplit_merge::init(vector<Cmomentum> &_particles, vector<Cmomentum> *protocones, double R2){
  // init particles
  if (init_particles(_particles) != 0)
    return 1;

  // browse protocones
  return add_protocones(protocones, R2);
}


// initialisation function for particle list
//  - _particles  list of particles
//-------------------------------------------------------------
int Csplit_merge::init_particles(vector<Cmomentum> &_particles){
  full_clear();

  // compute the list of particles
  // here, we do not need to throw away particles 
  // with infinite rapidity (colinear withthe beam)
  particles = _particles;
  n = particles.size();

  // set up the list of particles left.
  init_pleft();

  indices = new int[n];

  return 0;
}
  
// build initial list of remaining particles
//--------------------------------------
int Csplit_merge::init_pleft(){
  // at this level, we only rule out particles with 
  // infinite rapidity
  // in the parent particle list, index contain the run 
  // at which particles are puts in jets:
  //  - -1 means infinity rapidity
  //  -  0 means not included
  //  -  i mean included at run i
  int i,j;
  j=0;
  p_remain.clear();
  for (i=0;i<n;i++){
    // set ref for checkxor
    particles[i].ref.randomize();

    // check if rapidity is not infinite
    if ((particles[i].pz) != (particles[i].E)){
      p_remain.push_back(particles[i]);
      // left particles are marked with a 1
      p_remain[j].index = 1;
      // set up parent index for tracability
      p_remain[j].parent_index = i;
      j++;
      // set up parent-particle index
      particles[i].index = 0;
    } else {
      particles[i].index = -1;
    }
  }
  n_left = p_remain.size();

  n_pass = 0;

  return 0;
}

// partial clearance
// we want to keep
//   particle list and indices
// for future usage, so do not clear it !
// this is done in full_clear
//----------------------------------------
int Csplit_merge::partial_clear(){
  // release jets
  jets.clear();
  candidates.clear();
#ifdef ALLOW_MERGE_IDENTICAL_PROTOCONES
  if (merge_identical_protocones)
    cand_refs.clear();
#endif

  p_remain.clear();

  return 0;
}


// full clearance
//----------------
int Csplit_merge::full_clear(){
  partial_clear();

  // clear previously allocated memory
  if (indices != NULL){
    delete[] indices;
  }
  particles.clear();

  return 0;
}


// add a list of protocones
//  - _particles  list of particles
//  - protocones  list of protocones (initial jet candidates)
//  - R2          cone radius (squared)
//-------------------------------------------------------------
int Csplit_merge::add_protocones(vector<Cmomentum> *protocones, double R2){
  int i;
  Cmomentum *c;
  Cmomentum *v;
  double eta, phi;
  double dx, dy;
  Cjet jet;

  if (protocones->size()==0)
    return 1;

  // browse protocones
  // for each of them, build the list of particles in them
  for (vector<Cmomentum>::iterator p_it = protocones->begin();p_it != protocones->end();p_it++){
    // initialise variables
    c = &(*p_it);

    // note: cones have been tested => their (eta,phi) coordinates are computed
    eta = c->eta;
    phi = c->phi;

    // browse particles to create cone content
    // note that jet is always initialised with default values at this level
    jet.v = Cmomentum();
    jet.content.clear();
    for (i=0;i<n_left;i++){
      v = &(p_remain[i]);
      // for security, avoid to include particles with infinite rapidity)
      // NO NEEDED ANYMORE SINCE REMOVED FROM p_remain FROM THE BEGINNING
      //if (fabs(v->pz)!=v->E){
      dx = eta - v->eta;
      dy = fabs(phi - v->phi);
      if (dy>M_PI) 
	dy -= 2.0*M_PI;
      if (dx*dx+dy*dy<R2){
	jet.content.push_back(v->parent_index);
	jet.v+= *v;
	v->index=0;
      }
    }
    jet.n=jet.content.size();
    
    // add it to the list of jets
    // we maintain a set of checkxor to avoid having twice the
    // same content. This situation might lead to incorrect ehaviours
    // if the content is already present, insertion into cand_refs fails
#ifdef ALLOW_MERGE_IDENTICAL_PROTOCONES
    if ((merge_identical_protocones) && (cand_refs.insert(jet.v.ref).second))
#endif
      candidates.insert(jet);
  }
  
  // update list of included particles
  n_pass++;

  int j=0;
  for (i=0;i<n_left;i++){
    if (p_remain[i].index){
      // copy particle
      p_remain[j]=p_remain[i];
      p_remain[j].parent_index=p_remain[i].parent_index;      
      p_remain[j].index=0;
      j++;
      // set run in initial list
      particles[p_remain[i].parent_index].index = n_pass;
    }
  }
  n_left = j;
  p_remain.resize(j);

  return 0;
}

/// save the event on disk
///  - flux   stream used to save jet contents
int Csplit_merge::save_content(FILE *flux){
  jet_iterator it_j;
  Cjet *j1;
  int i1, i2;

  fprintf(flux, "%d jets found\n", (int) jets.size());
  for (it_j = jets.begin(), i1=0 ; it_j != jets.end() ; it_j++, i1++){
    j1 = &(*it_j);
    j1->v.build_etaphi();
    fprintf(flux, "%lf\t%lf\t%le\t%d\n", 
	    j1->v.eta, j1->v.phi, j1->v.perp(), j1->n);
  }
  
  for (it_j = jets.begin(), i1=0 ; it_j != jets.end() ; it_j++, i1++){
    j1 = &(*it_j);
    for (i2=0;i2<j1->n;i2++)
      fprintf(flux, "%lf\t%lf\t%d\t%d\n", 
	      particles[j1->content[i2]].eta, particles[j1->content[i2]].phi,
	      j1->content[i2], i1);
  }
  
  return 0;
}


/// show current jets/candidate status
int Csplit_merge::show(){
  jet_iterator it_j;
  cjet_iterator it_c;
  Cjet *j;
  const Cjet *c;
  int i1, i2;

  for (it_j = jets.begin(), i1=0 ; it_j != jets.end() ; it_j++, i1++){
    j = &(*it_j);
    fprintf(stdout, "jet %2d: %le\t%le\t%le\t%le\t", i1+1,
	    j->v.px, j->v.py, j->v.pz, j->v.E);
    for (i2=0;i2<j->n;i2++)
      fprintf(stdout, "%d ", j->content[i2]);
    fprintf(stdout, "\n");
  }
  
  for (it_c = candidates.begin(), i1=0 ; it_c != candidates.end() ; it_c++, i1++){
    c = &(*it_c);
    fprintf(stdout, "cdt %2d: %le\t%le\t%le\t%le\t", i1+1,
	    c->v.px, c->v.py, c->v.pz, c->v.E);
    for (i2=0;i2<c->n;i2++)
      fprintf(stdout, "%d ", c->content[i2]);
    fprintf(stdout, "\n");
  }
  
  fprintf(stdout, "\n");
  return 0;
}


/*
 * really do the splitting and merging
 * At the end, the vector jets is filled with the jets found.
 * the 'content' field of each jets contains the indices
 * of the particles included in that jet. 
 *  - overlap_tshold    threshold for splitting/merging transition
 *  - ptmin             minimal pT allowed for jets
 * return the number of jets is returned
 ************************************************************/
int Csplit_merge::perform(double overlap_tshold, double ptmin){
  // iterators for the 2 jets
  cjet_iterator j1;
  cjet_iterator j2;

  double ptmin2 = ptmin*ptmin;
  bool still_erasing;

  if (candidates.size()==0)
    return 0;

  // overlap
  Cmomentum overlap;

  // avoid to compute tshold*tshold at each overlap
  double overlap_tshold2 = overlap_tshold*overlap_tshold;

  do{
    // remove jets with too low a pT
    // note that we know that we have at least on element in the list
    still_erasing = true;
    do{
      j1 = candidates.end();
      j1--;
      if (j1->v.perp2()<ptmin2)
	candidates.erase(j1);
      else
	still_erasing = false;
    } while ((still_erasing) && (candidates.size()>0));

    if (candidates.size()>0){
      // browse for the first jet
      j1 = candidates.begin();
      
      // browse for the second jet
      j2 = j1;
      j2++;
      
      while (j2 != candidates.end()){
#ifdef DEBUG
	show();
#endif
	// check overlapping
	if (get_overlap(*j1, *j2, &overlap)){
	  // check if overlapping energy passes threshold
	  if (overlap.perp2() < overlap_tshold2*j2->v.perp2()){
	    // split jets
	    split(j1, j2);
	    
	    // update iterators
	    j2 = j1 = candidates.begin();
	  } else {
	    // merge jets
	    merge(j1, j2);
	    
	    // update iterators
	    j2 = j1 = candidates.begin();
	  }
	}
	j2++;
      } // end of loop on the second jet
      
      // all "second jet" passed without overlapping
      // (otherwise we won't leave the j2 loop)
      // => store jet 1 as real jet
      jets.push_back(*j1);
      jets[jets.size()-1].v.build_etaphi();
      jets[jets.size()-1].pass = particles[j1->content[0]].index;
#ifdef ALLOW_MERGE_IDENTICAL_PROTOCONES
      cand_refs.erase(j1->v.ref);
#endif
      candidates.erase(j1);
    }
  } while (candidates.size()>0);

  // sort jets by pT
  sort(jets.begin(), jets.end());
#ifdef DEBUG
      show();
#endif

  return jets.size();
}


// get the overlap between 2 jets
//  - j1   first jet
//  - j2   second jet
//  - v    returned overlap momentum
// return true if overlapping, false if disjoint
////////////////////////////////////////////////////////////////
bool Csplit_merge::get_overlap(const Cjet &j1, const Cjet &j2, Cmomentum *v){
  int i1,i2;
  bool is_overlap;

  // initialise
  i1=i2=idx_size=0;
  is_overlap = false;
  *v = Cmomentum();

  // compute overlap
  // at the same time, we store union in indices
  do{
    if (j1.content[i1]<j2.content[i2]){
      indices[idx_size] = j1.content[i1];
      i1++;
    } else if (j1.content[i1]>j2.content[i2]){
      indices[idx_size] = j2.content[i2];
      i2++;
    } else { // (j1.content[i1]==j2.content[i2])
      (*v) += particles[j1.content[i1]];
      indices[idx_size] = j1.content[i1];
      i1++;
      i2++;
      is_overlap = true;
    }
    idx_size++;
  } while ((i1<j1.n) && (i2<j2.n));

  // finish computing union
  // (only needed if overlap !)
  if (is_overlap){
    while (i1<j1.n){
      indices[idx_size] = j1.content[i1];
      i1++;
      idx_size++;
    }
    while (i2<j2.n){
      indices[idx_size] = j2.content[i2];
      i2++;
      idx_size++;
    }
  }

  return is_overlap;
}


// split the two given jet.
// during this procedure, the jets j1 & j2 are replaced
// by 2 new jets. Common particles are associted to the 
// closest initial jet.
//  - it_j1  iterator of the first jet in 'candidates'
//  - it_j2  iterator of the second jet in 'candidates'
//  - j1     first jet (Cjet instance)
//  - j2     second jet (Cjet instance)
// return true on success, false on error
////////////////////////////////////////////////////////////////
bool Csplit_merge::split(cjet_iterator &it_j1, cjet_iterator &it_j2){
  int i1, i2;
  Cjet jet1, jet2;
  double eta1, phi1, eta2, phi2;
  double dx1, dy1, dx2, dy2;
  Cmomentum tmp;
  Cmomentum *v;

  // shorthand to avoid having "->" all over the place
  const Cjet & j1 = * it_j1;
  const Cjet & j2 = * it_j2;

  i1=i2=0;
  jet2.v = jet1.v = Cmomentum();

  // copmpute centroids
  tmp = j1.v;
  tmp.build_etaphi();
  eta1 = tmp.eta;
  phi1 = tmp.phi;

  tmp = j2.v;
  tmp.build_etaphi();
  eta2 = tmp.eta;
  phi2 = tmp.phi;

  jet1.v = jet2.v = Cmomentum();  

  // compute jet splitting
  do{
    if (j1.content[i1]<j2.content[i2]){
      // particle i1 belong only to jet 1
      jet1.content.push_back(j1.content[i1]);
      jet1.v += particles[j1.content[i1]];
      i1++;
    } else if (j1.content[i1]>j2.content[i2]){
      // particle i2 belong only to jet 2
      jet2.content.push_back(j2.content[i2]);
      jet2.v += particles[j2.content[i2]];
      i2++;
    } else { // (j1.content[i1]==j2.content[i2])
      // common particle, decide which is the closest centre
      v = &(particles[j1.content[i1]]);

      // distance w.r.t. centroid 1
      dx1 = eta1 - v->eta;
      dy1 = fabs(phi1 - v->phi);
      if (dy1>M_PI) 
	dy1 -= 2.0*M_PI;

      // distance w.r.t. centroid 2
      dx2 = eta2 - v->eta;
      dy2 = fabs(phi2 - v->phi);
      if (dy2>M_PI) 
	dy2 -= 2.0*M_PI;

      //? what when == ?
      if (dx1*dx1+dy1*dy1<dx2*dx2+dy2*dy2){
	// particle i1 belong only to jet 1
	jet1.content.push_back(j1.content[i1]);
	jet1.v += *v;
      } else {
	// particle i2 belong only to jet 2
	jet2.content.push_back(j2.content[i2]);
	jet2.v += *v;
      }      

      i1++;
      i2++;
    }
  } while ((i1<j1.n) && (i2<j2.n));
  
  while (i1<j1.n){
    jet1.content.push_back(j1.content[i1]);
    jet1.v += particles[j1.content[i1]];
    i1++;
  }
  while (i2<j2.n){
    jet2.content.push_back(j2.content[i2]);
    jet2.v += particles[j2.content[i2]];
    i2++;
  }

  // finalise jets
  jet1.n = jet1.content.size();
  jet2.n = jet2.content.size();

  // remove previous jets
#ifdef ALLOW_MERGE_IDENTICAL_PROTOCONES
  cand_refs.erase(j1.v.ref);
  cand_refs.erase(j2.v.ref);
#endif
  candidates.erase(it_j1);
  candidates.erase(it_j2);

  // reinsert new ones
  // we check that the candidates were not already present
  // by checking insertion in the checkxor cand_refs set
#ifdef ALLOW_MERGE_IDENTICAL_PROTOCONES
  if ((merge_identical_protocones) && (cand_refs.insert(jet1.v.ref).second))
#endif
    candidates.insert(jet1);
#ifdef ALLOW_MERGE_IDENTICAL_PROTOCONES
  if ((merge_identical_protocones) && (cand_refs.insert(jet2.v.ref).second))
#endif
    candidates.insert(jet2);

  return true;
}

// merge the two given jet.
// during this procedure, the jets j1 & j2 are replaced
// by 1 single jets containing both of them.
//  - it_j1  iterator of the first jet in 'candidates'
//  - it_j2  iterator of the second jet in 'candidates'
// return true on success, false on error
////////////////////////////////////////////////////////////////
bool Csplit_merge::merge(cjet_iterator &it_j1, cjet_iterator &it_j2){
  Cjet jet;
  int i;

  // build new jet
  // note: particles within j1 & j2 have already been stored in indices
  for (i=0;i<idx_size;i++){
    jet.content.push_back(indices[i]);
    jet.v += particles[indices[i]];
  }
  jet.n = jet.content.size();

  // remove old candidates
#ifdef ALLOW_MERGE_IDENTICAL_PROTOCONES
  if (merge_identical_protocones){
    cand_refs.erase(it_j1->v.ref);
    cand_refs.erase(it_j2->v.ref);
  }
#endif
  candidates.erase(it_j1);
  candidates.erase(it_j2);

  // reinsert new candidate
  // we check that the candidates were not already present
  // by checking insertion in the checkxor cand_refs set
#ifdef ALLOW_MERGE_IDENTICAL_PROTOCONES
  if ((merge_identical_protocones) && (cand_refs.insert(jet.v.ref).second))
#endif
    candidates.insert(jet);

  return true;
}

}
